import unittest

from pyravendb.tests.test_base import TestBase


class TestGet(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestGet, cls).setUpClass()
        cls.db.put("products/101", {"Name": "test"}, {})
        cls.db.put("products/10", {"Name": "test"}, {})
        cls.response = cls.db.get("products/101")
        cls.other_response = cls.db.get("products/10")

    def test_equal(self):
        self.assertEqual(self.response["Results"][0]["@metadata"]["@id"], "products/101")

    def test_not_equal(self):
        self.assertNotEqual(self.response["Results"][0]["@metadata"]["@id"],
                            self.other_response["Results"][0]["@metadata"]["@id"])

    def test_null(self):
        self.assertIsNone(self.db.get("product")["Results"][0])


if __name__ == "__main__":
    unittest.main()



import unittest

from pyravendb.d_commands import commands_data
from pyravendb.data.patches import ScriptedPatchRequest
from pyravendb.tests.test_base import TestBase


class TestDelete(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestDelete, cls).setUpClass()
        cls.put_command1 = commands_data.PutCommandData("products/999",
                                                        document={"Name": "tests", "Category": "testing"},
                                                        metadata={"Raven-Python-Type": "Products"})
        cls.put_command2 = commands_data.PutCommandData("products/1000",
                                                        document={"Name": "testsDelete", "Category": "testing"},
                                                        metadata={"Raven-Python-Type": "Products"})
        cls.delete_command = commands_data.DeleteCommandData("products/1000")
        patch = ScriptedPatchRequest("this.Name = 'testing';")
        cls.scripted_patch_command = commands_data.ScriptedPatchCommandData("products/999", patch)

    def test_success_one_command(self):
        batch_result = self.db.batch([self.put_command1])
        self.assertEqual(len(batch_result), 1)

    def test_success_multi_commands(self):
        batch_result = self.db.batch([self.put_command1, self.put_command2, self.delete_command])
        self.assertEqual(len(batch_result), 3)

    def test_scripted_patch(self):
        self.db.batch([self.put_command1, self.scripted_patch_command])
        result = self.db.get("products/999")
        self.assertEqual(result["Results"][0]["Name"], "testing")

    def test_fail(self):
        with self.assertRaises(ValueError):
            self.db.batch([self.put_command1, self.put_command2, None])

    if __name__ == "__main__":
        unittest.main()



class DatabaseDocument(object):
    __slots__ = ["database_id", "settings", "secured_settings", "disabled"]

    def __init__(self, database_id, settings=None, secured_settings=None, disabled=False):
        if secured_settings is None:
            secured_settings = {}
        if settings is None:
            settings = {}
        self.database_id = database_id
        self.settings = settings
        self.secured_settings = secured_settings
        self.disabled = disabled

    def to_json(self):
        return {"Disabled": self.disabled, "SecuredSettings": self.secured_settings, "Settings": self.settings}


class ReplicationDocument(object):
    def __init__(self, destinations):
        self.Id = "Raven/Replication/Destinations"
        self.destinations = []
        self.add_destinations(destinations)
        self.source = None

    def add_destinations(self, destinations):
        if not isinstance(destinations, list):
            destinations = [destinations]
        self.destinations = [destination.to_json() for destination in destinations]


class ReplicationDestination(object):
    __slots__ = ["destination", "url", "database", "user_name", "password", "domain", "api_key", "disabled",
                 "_humane"]

    def __init__(self, url, database, user_name=None, password=None, domain=None, api_key=None, disabled=False,
                 client_visible_url=None):
        self._humane = None if url is None else "{0} {1}".format(url, database)
        self.destination = {"url": url[:-1] if url.endswith('/') else url, "database": database, "username": user_name,
                            "password": password,
                            "domain": domain, "ApiKey": api_key,
                            "disabled": disabled, "clientVisibleUrl": client_visible_url,
                            "humane": None if url is None else "{0} {1}".format(url, database)}

    def to_json(self):
        return self.destination



class _CommandData(object):
    def __init__(self, key, etag=None, metadata=None):
        self.key = key
        self.metadata = metadata
        self.etag = etag
        self.additionalData = None
        self.__command = True

    def to_json(self):
        raise NotImplementedError

    @property
    def command(self):
        return self.__command


class PutCommandData(_CommandData):
    def __init__(self, key, etag=None, document=None, metadata=None):
        super(PutCommandData, self).__init__(key, etag, metadata)
        self._method = "PUT"
        self.document = document

    def to_json(self):
        data = {"Method": self._method, "Key": self.key, "Document": self.document,
                "Metadata": self.metadata, "AdditionalData": self.additionalData}
        if self.etag is not None:
            data["Etag"] = self.etag

        return data


class DeleteCommandData(_CommandData):
    def __init__(self, key, etag=None):
        super(DeleteCommandData, self).__init__(key, etag)
        self._method = "DELETE"
        self.additionalData = None

    def to_json(self):
        return {"Key": self.key, "Etag": self.etag, "Method": self._method, "AdditionalData": self.additionalData}


class ScriptedPatchCommandData(_CommandData):
    def __init__(self, key, scripted_patch, etag=None, metadata=None, patch_if_missing=None, additional_data=None,
                 debug_mode=False):
        """
        @param key: key of a document to patch.
        :type str
        @param scripted_patch: (using JavaScript) that is used to patch the document.
        :type ScriptedPatchRequest
        @param etag: Current document etag, used for concurrency checks (null to skip check).
        :type str
        @param metadata: representing document's metadata.
        :type dict
        @param patch_if_missing: (using JavaScript) that is used to patch a default document.
        :type ScriptedPatchRequest
        @param additional_data: Additional command data. For internal use only.
        :type dict
        @param debug_mode: Indicates in the operation should be run in debug mode.
                           for additional information in response.
        :type bool

        """

        super(ScriptedPatchCommandData, self).__init__(key, etag)
        if additional_data is None:
            additional_data = {}
        if metadata is None:
            metadata = {}
        self._method = "EVAL"
        self.scripted_patch = scripted_patch
        self.metadata = metadata
        self.patch_if_missing = patch_if_missing
        self.additional_data = additional_data
        self.debug_Mode = debug_mode

    def to_json(self):
        data = {"Key": self.key, "Method": self._method, "Patch": self.scripted_patch.to_json(),
                "DebugMode": self.debug_Mode, "AdditionalData": self.additionalData, "Metadata": self.metadata}
        if self.etag is not None:
            data["Etag"] = self.etag
        if self.patch_if_missing is not None:
            data["PatchIfMissing"] = self.patch_if_missing.to_json()

        return data



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import documentstore
from pyravendb.store.session_query import QueryOperator
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexDefinition, SortOptions
import unittest


class Product(object):
    def __init__(self, name, key, order):
        self.name = name
        self.key = key
        self.order = order


class Company(object):
    def __init__(self, name, product):
        self.name = name
        self.product = product


class TestQuery(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestQuery, cls).setUpClass()
        cls.index_map = ("from doc in docs "
                         "select new {"
                         "name = doc.name,"
                         "key = doc.key,"
                         "doc_id = (doc.name + \" \") + doc.key}")
        cls.index_sort = IndexDefinition(index_map=cls.index_map, sort_options={"key": SortOptions.float},
                                         stores={"doc_id": "Yes"})
        cls.db.put_index("Testing_Sort", index_def=cls.index_sort, overwrite=True)
        cls.db.put("products/101", {"name": "test101", "key": 2, "order": "a"},
                   {"Raven-Entity-Name": "Products", "Raven-Python-Type": Product.__module__ + ".Product"})
        cls.db.put("products/10", {"name": "test10", "key": 3, "order": "b"}, {})
        cls.db.put("products/106", {"name": "test106", "key": 4, "order": "c"}, {})
        cls.db.put("products/107", {"name": "test107", "key": 5, "order": None},
                   {"Raven-Entity-Name": "Products", "Raven-Python-Type": Product.__module__ + ".Product"})
        cls.db.put("products/103", {"name": "test107", "key": 6},
                   {"Raven-Entity-Name": "Products", "Raven-Python-Type": Product.__module__ + ".Product"})
        cls.db.put("products/108", {"name": "new_testing", "key": 90, "order": "d"},
                   {"Raven-Entity-Name": "Products", "Raven-Python-Type": Product.__module__ + ".Product"})
        cls.db.put("orders/105", {"name": "testing_order", "key": 92, "product": "products/108"},
                   {"Raven-Entity-Name": "Orders"})
        cls.db.put("company/1",
                   {"name": "withNesting", "product": {"name": "testing_order", "key": 4, "order": None}},
                   {"Raven-Entity-Name": "Companies"})
        cls.db.put("specials/1", {"url": "https://ravendb.net/"})
        cls.db.put("specials/2",
                   {"url": "http://www.example.com/ebx/LinkResolverServlet?classofcontent=Standard&id=63935",
                    "date": "2017.03.05T00:00:00.000"})
        cls.db.put("specials/3",
                   {"wild_cards": "%~#+!$,;'*[]?*",
                    "date": "2017.03.05T00:00:01.000"})
        cls.document_store = documentstore(cls.default_url, cls.default_database)
        cls.document_store.initialize()

    def test_where_equal_dynamic_index(self):
        with self.document_store.open_session() as session:
            query_results = list(session.query().where_equals("name", "test101"))
            self.assertEqual(query_results[0].name, "test101")

    def test_where_equal_double(self):
        with self.document_store.open_session() as session:
            query_results = list(session.query().where_equals("name", "test101").where_equals("key", 4))
            self.assertEqual(len(query_results), 2)

    def test_where_equal_double_and_operator(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(object_type=Product, using_default_operator=QueryOperator.AND).where_equals(
                    "name", "test107").where_equals("key", 5))
            self.assertEqual(len(query_results), 1)

    def test_where_in(self):
        with self.document_store.open_session() as session:
            query_result = list(session.query().where_in("name", ["test101", "test107", "test106"]))
            self.assertEqual(len(query_result), 4)

    def test_where_starts_with(self):
        with self.document_store.open_session() as session:
            query_result = list(session.query(Product).where_starts_with("name", "n"))
            self.assertEqual(query_result[0].name, "new_testing")

    def test_where_ends_with(self):
        with self.document_store.open_session() as session:
            query_result = list(session.query(Product).where_ends_with("name", "7"))
            self.assertEqual(query_result[0].name, "test107")

    def test_query_fail_with_index(self):
        with self.document_store.open_session() as session:
            with self.assertRaises(exceptions.ErrorResponseException):
                list(session.query(index_name="s").where(Tag="Products"))

    def test_query_with_special_characters(self):
        with self.document_store.open_session() as session:
            url = "https://ravendb.net/"
            query_result = list(session.query().where_equals("url", url).select("url"))
            self.assertTrue(len(query_result) > 0)
            self.assertTrue(query_result[0].url == url)

    def test_query_with_wild_cards_and_select(self):
        with self.document_store.open_session() as session:
            url = "http://www.example.com/ebx/LinkResolverServlet?classofcontent=Standard&id=63935"
            query_result = list(session.query().where_equals("url", url).select("date"))
            self.assertTrue(len(query_result) > 0)
            self.assertIsNotNone(query_result[0].date)
            self.assertFalse(hasattr(query_result[0], "url"))

    def test_query_with_only_wild_cards(self):
        with self.document_store.open_session() as session:
            wild_cards = "%~#+!$,;'*[]?*"
            query_result = list(session.query().where_equals("wild_cards", wild_cards))
            self.assertTrue(len(query_result) > 0)
            self.assertIsNotNone(query_result[0].wild_cards)

    def test_query_success_with_where(self):
        with self.document_store.open_session() as session:
            query_results = list(session.query().where(name="test101", key=[4, 6, 90]))
            self.assertIsNotNone(query_results)

    def test_query_success_with_index(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where(key=[4, 6, 90]))
            self.assertEqual(len(query_results), 3)

    def test_where_between(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where_between("key", 2, 4))
            self.assertEqual(len(query_results), 1)

    def test_where_between_or_equal(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where_between_or_equal("key",
                                                                                                                 2, 4))
            self.assertEqual(len(query_results), 3)

    def test_query_with_order_by(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(wait_for_non_stale_results=True).where_not_none("order").order_by("order"))
            self.assertEqual(query_results[0].order, "a")

    def test_query_with_order_by_descending(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(wait_for_non_stale_results=True).where_not_none("order").order_by_descending("order"))
            self.assertEqual(query_results[0].order, "d")

    def test_where_not_None(self):
        found_none = False
        with self.document_store.open_session() as session:
            query_results = list(session.query(wait_for_non_stale_results=True).where_not_none("order"))
            for result in query_results:
                if result.order is None:
                    found_none = True
                    break
            self.assertFalse(found_none)

    def test_where_with_include(self):
        with self.document_store.open_session() as session:
            list(session.query(wait_for_non_stale_results=True, includes="product").where(key=92))
            session.load("products/108")
        self.assertEqual(session.number_of_requests_in_session, 1)

    def test_query_with_nested_object(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(object_type=Company, nested_object_types={"product": Product}).where_equals(
                    "name", "withNesting"))
            self.assertTrue(isinstance(query_results[0], Company) and isinstance(query_results[0].product, Product))

    def test_query_with_fetch_terms(self):
        with self.document_store.open_session() as session:
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where_between_or_equal
                ("key", 2, 4).select("doc_id"))

            found_in_all = True
            for result in query_results:
                if not hasattr(result, "doc_id"):
                    found_in_all = False
                    break

            self.assertTrue(found_in_all)

    class Product:
        def __init__(self, name, key, order):
            self.name = name
            self.key = key
            self.order = order

    def test_stream(self):
        with self.document_store.open_session() as session:
            session.save_changes()
            list(session.query(wait_for_non_stale_results=True, index_name="Testing_Sort"))

        with self.document_store.open_session() as session:
            query_results = session.query(index_name="Testing_Sort")
            results = session.advanced.stream(query_results)
            count = 0
            for _ in results:
                count += 1
            self.assertEqual(count, 8)

    def test_stream_with_object_type(self):
        object_type = Company
        nested_object_types = {"product": Product}
        with self.document_store.open_session() as session:
            list(session.query(object_type=object_type, nested_object_types=nested_object_types).where_equals(
                "name", "withNesting"))

        with self.document_store.open_session() as session:
            query_results = session.query(index_name="Auto/Companies/Byname", ).where_equals("name", "withNesting")
            results = session.advanced.stream(query_results, object_type=object_type,
                                              nested_object_types=nested_object_types)
            for result in results:
                self.assertTrue(isinstance(result["document"], Company))


def test_where_starts_with_skip_take(self):
    with self.document_store.open_session() as session:
        query_result = list(session.query().where_starts_with("name", "test").take(4).skip("2"))
        self.assertEqual(len(query_result), 4)


if __name__ == "__main__":
    unittest.main()



import unittest

from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexQuery
from pyravendb.data.patches import ScriptedPatchRequest
from pyravendb.tests.test_base import TestBase
from pyravendb.data.operations import Operations
from pyravendb.data.indexes import IndexDefinition

class TestByIndexActions(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestByIndexActions, cls).setUpClass()
        cls.patch = ScriptedPatchRequest("this.Name = 'testing';")
        cls.db.put("testing/1", {"Name": "test"}, {"Raven-Entity-Name": "Testing"})
        cls.operation = Operations(cls.request_handler)
        index_map =("from doc in docs.Testing "
                    "select new{"
                    "Name = doc.Name}")
        index = IndexDefinition(index_map)
        cls.db.put_index("indexOperation", index, True)

    def test_update_by_index_success(self):
        result = self.db.update_by_index("indexOperation", IndexQuery("Name:test"), self.patch)
        status = self.operation.wait_for_operation_complete(result["OperationId"])
        self.assertTrue(status["Completed"] == True and status["Faulted"] == False)



    def test_update_by_index_fail(self):
        with self.assertRaises(exceptions.InvalidOperationException):
            result = self.db.update_by_index("c", IndexQuery("Name:test"), self.patch)
            self.operation.wait_for_operation_complete(result["OperationId"])

    def test_delete_by_index_fail(self):
        with self.assertRaises(exceptions.ErrorResponseException):
            self.db.delete_by_index("region2", IndexQuery("Name:Western"))

    def test_delete_by_index_success(self):
        result = self.db.delete_by_index("indexOperation", IndexQuery("Name:test"))
        status = self.operation.wait_for_operation_complete(result["OperationId"])
        self.assertTrue(status["Completed"] == True and status["Faulted"] == False)


if __name__ == "__main__":
    unittest.main()



class PKCS7Encoder(object):
    """
    Technique for padding a string as defined in RFC 2315, section 10.3,
    note #2
    """

    class InvalidBlockSizeError(Exception):
        """Raised for invalid block sizes"""
        pass

    def __init__(self, block_size=16):
        if block_size < 2 or block_size > 255:
            raise PKCS7Encoder.InvalidBlockSizeError('The block size must be ' \
                                                     'between 2 and 255, inclusive')
        self.block_size = block_size

    def encode(self, text):
        text_length = len(text)
        amount_to_pad = self.block_size - (text_length % self.block_size)
        if amount_to_pad == 0:
            amount_to_pad = self.block_size
        pad = chr(amount_to_pad)
        return text + pad * amount_to_pad

    def decode(self, text):
        pad = ord(text[-1])
        return text[:-pad]



from pyravendb.data.document_convention import DocumentConvention, Failover
from pyravendb.tools.indexqueue import IndexQueue
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from Crypto.Util.number import bytes_to_long
from pyravendb.custom_exceptions import exceptions
from pyravendb.tools.pkcs7 import PKCS7Encoder
import tempfile
import requests
import sys
import json
import hashlib
import base64
import os
from pyravendb.tools.utils import Utils
from threading import Timer, Lock


class HttpRequestsFactory(object):
    def __init__(self, url, database, convention=None, api_key=None, force_get_topology=False):
        self.url = url
        self._primary_url = url
        self.database = database
        self._primary_database = database
        self.primary = False
        self.api_key = api_key
        self.version_info = sys.version_info.major
        self.convention = convention
        if self.convention is None:
            self.convention = DocumentConvention()
        self.headers = {"Accept": "application/json", "Has-Api-key": 'true' if self.api_key is not None else 'false',
                        "Raven-Client-Version": "3.5.0.0"}
        self.replication_topology = IndexQueue()
        self.topology_change_counter = 0
        self.lock = Lock()
        self.topology = None
        self.request_count = 0
        self.force_get_topology = force_get_topology
        self._token = None
        self._current_api_key = None
        self._current_database = None

    def http_request_handler(self, path, method, data=None, headers=None, admin=False, force_read_from_master=False,
                             uri="databases", stream=False):
        if self.force_get_topology:
            self.force_get_topology = False
            self.get_replication_topology()

        if self.url == self._primary_url and self.database == self._primary_database and not self.primary:
            self.primary = True

        if self.database.lower() == self.convention.system_database:
            force_read_from_master = True
            uri = self.convention.system_database

        return self._execute_with_replication(path, method, headers=headers, data=data, admin=admin,
                                              force_read_from_master=force_read_from_master, uri=uri, stream=stream)

    def _execute_with_replication(self, path, method, headers, data=None, admin=False,
                                  force_read_from_master=False, uri="databases", stream=False):
        second_api_key = None
        while True:
            index = None
            url = None
            if not force_read_from_master:
                if admin:
                    url = "{0}/admin/{1}".format(self._primary_url, path)
                    second_api_key = self.api_key
                else:
                    if method == "GET":
                        if (path == "replication/topology" or "Hilo" in path) and not self.primary:
                            raise exceptions.InvalidOperationException(
                                "Cant get access to {0} when {1}(primary) is Down".format(path, self._primary_database))
                        elif self.convention.failover_behavior == Failover.read_from_all_servers:
                            with self.lock:
                                self.request_count += 1
                                index = self.request_count % (len(self.replication_topology) + 1)
                            if index != 0 or not self.primary:  # if 0 use the primary
                                index -= 1
                                destination = self.replication_topology.peek(index if index > 0 else 0)
                                url = "{0}/{1}/{2}/{3}".format(destination["url"], uri, destination["database"], path)
                                second_api_key = destination["credentials"].get("api_key", None)
                        elif not self.primary:
                            url = "{0}/{1}/{2}/{3}".format(self.url, uri, self.database, path)

                    else:
                        if not self.primary:
                            if self.convention.failover_behavior == \
                                    Failover.allow_reads_from_secondaries_and_writes_to_secondaries:
                                url = "{0}/{1}/{2}/{3}".format(self.url, uri, self.database, path)
                            else:
                                raise exceptions.InvalidOperationException(
                                    "Cant write to server when the primary is down when failover = {0}".format(
                                        self.convention.failover_behavior.name))

            if url is None:
                if not self.primary:
                    raise exceptions.InvalidOperationException(
                        "Cant read or write to the master because {0} is down".format(self._primary_database))
                url = "{0}/{1}/{2}/{3}".format(self._primary_url, uri, self._primary_database, path)
                if uri != "databases":
                    url = "{0}/{1}".format(self._primary_url, path)
                second_api_key = self.api_key
            with requests.session() as session:
                if headers is None:
                    headers = {}
                headers.update(self.headers)
                data = json.dumps(data, default=self.convention.json_default_method)
                response = session.request(method, url=url, data=data, headers=headers, stream=stream)
                if response.status_code == 412 or response.status_code == 401:
                    try:
                        oauth_source = response.headers.__getitem__("OAuth-Source")
                    except KeyError:
                        raise exceptions.InvalidOperationException(
                            "Something is not right please check your server settings (do you use the right api_key)")
                    self.do_auth_request(self.api_key, oauth_source, second_api_key)
                    continue
                if (response.status_code == 503 or response.status_code == 502) and \
                        not self.replication_topology.empty() and not (
                        path == "replication/topology" or "Hilo" in path):
                    if self.primary:
                        if self.convention.failover_behavior == Failover.fail_immediately or force_read_from_master:
                            raise exceptions.ErrorResponseException("Failed to get response from server")
                        self.primary = False
                        self.is_alive({"url": self._primary_url, "database": self._primary_database}, primary=True)

                    else:
                        with self.lock:
                            if not index:
                                index = 0
                            peek_item = self.replication_topology.peek(index)
                            if self.url == peek_item["url"] and self.database == peek_item["database"]:
                                self.is_alive(self.replication_topology.get(index))
                    if self.replication_topology.empty():
                        raise exceptions.ErrorResponseException("Please check your databases")
                    destination = self.replication_topology.peek()
                    self.database = destination["database"]
                    self.url = destination["url"]
                    second_api_key = destination["credentials"].get("api_key", None)
                    continue
            return response

    def is_alive(self, destination, primary=False):
        with requests.session() as session:
            while True:
                response = session.request("GET",
                                           "{0}/databases/{1}/replication/topology?check-server-reachable".format(
                                               destination["url"], destination["database"]), headers=self.headers)
                if response.status_code == 412 or response.status_code == 401:
                    try:
                        try:
                            oauth_source = response.headers.__getitem__("OAuth-Source")
                        except KeyError:
                            raise exceptions.InvalidOperationException(
                                "Something is not right please check your server settings")
                        self.do_auth_request(self.api_key, oauth_source)
                    except exceptions.ErrorResponseException:
                        break
                if response.status_code == 200:
                    if primary:
                        self.primary = True
                    else:
                        self.replication_topology.put(destination)
                    return
                else:
                    break
            is_alive_timer = Timer(5, lambda: self.is_alive(destination, primary))
            is_alive_timer.daemon = True
            is_alive_timer.start()

    def check_database_exists(self, path):
        return self.http_request_handler(path, "GET", force_read_from_master=True, uri="docs")

    def call_hilo(self, type_tag_name, max_id, etag):
        headers = {"if-None-Match": "\"" + etag + "\""}
        put_url = "docs/Raven%2FHilo%2F{0}".format(type_tag_name)
        response = self.http_request_handler(put_url, "PUT", data={"Max": max_id},
                                             headers=headers)
        if response.status_code == 409:
            raise exceptions.FetchConcurrencyException(response.json["Error"])
        if response.status_code != 201:
            raise exceptions.ErrorResponseException("Something is wrong with the request")

    def update_replication(self, topology_file):
        with open(topology_file, 'w+') as f:
            f.write(json.dumps(self.topology))
            self.replication_topology.queue.clear()
            self._load_topology()

    def check_replication_change(self, topology_file):

        try:
            response = self.http_request_handler("replication/topology", "GET")
            if response.status_code == 200:
                topology = response.json()
                with self.lock:
                    if self.topology != topology:
                        self.topology = topology
                        self.update_replication(topology_file)
            elif response.status_code != 400 and response.status_code != 404 and not self.topology:
                raise exceptions.ErrorResponseException(
                    "Could not connect to the database {0} please check the problem".format(self._primary_database))
        except exceptions.InvalidOperationException:
            pass
        if not self.database.lower() == self.convention.system_database:
            timer = Timer(60 * 5, lambda: self.check_replication_change(topology_file))
            timer.daemon = True
            timer.start()

    def get_replication_topology(self):
        with self.lock:
            hash_name = hashlib.md5(
                "{0}/{1}".format(self._primary_url, self._primary_database).encode('utf-8')).hexdigest()
            topology_file = "{0}{1}RavenDB_Replication_Information_For - {2}".format(tempfile.gettempdir(), os.path.sep,
                                                                                     hash_name)
            try:
                with open(topology_file, 'r') as f:
                    self.topology = json.loads(f.read())
                    self._load_topology()
            except IOError:
                pass

        self.check_replication_change(topology_file)

    def _load_topology(self):
        for destination in self.topology["Destinations"]:
            if not destination["Disabled"] and not destination["IgnoredClient"]:
                self.replication_topology.put({"url": destination["Url"], "database": destination["Database"],
                                               "credentials": {"api_key": destination["ApiKey"],
                                                               "domain": destination["Domain"]}})

    def do_auth_request(self, api_key, oauth_source, second_api_key=None):
        api_name, secret = api_key.split('/', 1)
        tries = 1
        headers = {"grant_type": "client_credentials"}
        data = None
        with requests.session() as session:
            while True:
                oath = session.request(method="POST", url=oauth_source,
                                       headers=headers, data=data)
                if oath.reason == "Precondition Failed":
                    if tries > 1:
                        if not (second_api_key and self.api_key != second_api_key and tries < 3):
                            raise exceptions.ErrorResponseException("Unauthorized")
                        api_name, secret = second_api_key.split('/', 1)
                        tries += 1

                    authenticate = oath.headers.__getitem__("www-authenticate")[len("Raven  "):]
                    challenge_dict = dict(item.split("=", 1) for item in authenticate.split(','))

                    exponent_str = challenge_dict.get("exponent", None)
                    modulus_str = challenge_dict.get("modulus", None)
                    challenge = challenge_dict.get("challenge", None)

                    exponent = bytes_to_long(base64.standard_b64decode(exponent_str))
                    modulus = bytes_to_long(base64.standard_b64decode(modulus_str))

                    rsa = RSA.construct((modulus, exponent))
                    cipher = PKCS1_OAEP.new(rsa)

                    iv = get_random_bytes(16)
                    key = get_random_bytes(32)
                    encoder = PKCS7Encoder()

                    cipher_text = cipher.encrypt(key + iv)
                    results = []
                    results.extend(cipher_text)

                    aes = AES.new(key, AES.MODE_CBC, iv)
                    sub_data = Utils.dict_to_string({"api key name": api_name, "challenge": challenge,
                                                     "response": base64.b64encode(hashlib.sha1(
                                                         '{0};{1}'.format(challenge, secret).encode(
                                                             'utf-8')).digest())})

                    results.extend(aes.encrypt(encoder.encode(sub_data)))
                    data = Utils.dict_to_string({"exponent": exponent_str, "modulus": modulus_str,
                                                 "data": base64.standard_b64encode(bytearray(results))})

                    if exponent is None or modulus is None or challenge is None:
                        raise exceptions.InvalidOperationException(
                            "Invalid response from server, could not parse raven authentication information:{0} ".format(
                                authenticate))
                    tries += 1
                elif oath.status_code == 200:
                    oath_json = oath.json()
                    body = oath_json["Body"]
                    signature = oath_json["Signature"]
                    if not sys.version_info.major > 2:
                        body = body.encode('utf-8')
                        signature = signature.encode('utf-8')
                    with self.lock:
                        self._token = "Bearer {0}".format(
                            {"Body": body, "Signature": signature})
                        self.headers.update({"Authorization": self._token})
                    break
                else:
                    raise exceptions.ErrorResponseException(oath.reason)



from pyravendb.tests.test_base import TestBase
from pyravendb.data.indexes import IndexDefinition
from pyravendb.data.indexes import FieldIndexing
from pyravendb.store.document_store import documentstore
from datetime import datetime


class LastFm(object):
    def __init__(self, artist="", track_id="", title="", datetime_time=None, tags=None):
        self.artist = artist
        self.track_id = track_id
        self.title = title
        self.datetime_time = datetime_time if datetime_time is not None else datetime.now()
        self.tags = tags if tags is not None else []


class LastFmAnalyzed(object):
    class Result(object):
        def __init__(self, query):
            self.query = query

    def __init__(self):
        index_map = ("from song in docs.LastFms "
                     "select new {"
                     "query = new object[] {"
                     "song.artist,"
                     "((object)song.datetime_time),"
                     "song.tags,"
                     "song.title,"
                     "song.track_id}}")

        indexes = {"query": FieldIndexing.analyzed}

        self.index_definition = IndexDefinition(index_map=index_map, indexes=indexes)

    def execute(self, store):
        store.database_commands.put_index(LastFmAnalyzed.__name__, self.index_definition, True)


class FullTextSearchTest(TestBase):
    @classmethod
    def setUpClass(cls):
        super(FullTextSearchTest, cls).setUpClass()
        cls.db.put("LastFms/1", {"artist": "Tania Maria", "track_id": "TRALPJJ128F9311763", "title": "Come With Me",
                                 "datetime_time": datetime.now(), "tags": None},
                   {"Raven-Entity-Name": "LastFms", "Raven-Python-Type": "full_text_search_test.LastFm"})
        cls.db.put("LastFms/2", {"artist": "Meghan Trainor", "track_id": "TRBCNGI128F42597B4", "title": "Me Too",
                                 "datetime_time": datetime.now(), "tags": None},
                   {"Raven-Entity-Name": "LastFms", "Raven-Python-Type": "full_text_search_test.LastFm"})
        cls.db.put("LastFms/3", {"artist": "Willie Bobo", "track_id": "TRAACNS128F14A2DF5", "title": "Spanish Grease",
                                 "datetime_time": datetime.now(), "tags": None},
                   {"Raven-Entity-Name": "LastFms", "Raven-Python-Type": "full_text_search_test.LastFm"})
        cls.document_store = documentstore(cls.default_url, cls.default_database)
        cls.document_store.initialize()
        LastFmAnalyzed().execute(cls.document_store)

    def test_full_text_search_one(self):
        with self.document_store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me"))
            self.assertTrue(str(query[0].title) == "Come With Me" and str(query[1].title) == "Me Too")

    def test_full_text_search_two(self):
        with self.document_store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me").search(
                "query", search_terms="Bobo"))
            self.assertEqual(len(query), 3)

    def test_full_text_search_with_boost(self):
        with self.document_store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me").boost(10).
                         search("query", search_terms="Bobo").boost(2))
            self.assertTrue(str(query[0].title) == "Come With Me" and str(query[1].title) == "Me Too" and str(
                query[2].title) == "Spanish Grease")



import unittest

from pyravendb.custom_exceptions.exceptions import DocumentDoesNotExistsException
from pyravendb.data.patches import ScriptedPatchRequest
from pyravendb.tests.test_base import TestBase


class TestPatch(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestPatch, cls).setUpClass()
        cls.db.put("products/10", {"Name": "test"}, {})

    def test_patch_success_ignore_missing(self):
        self.assertIsNotNone(self.db.patch("products/10", ScriptedPatchRequest("this.Name = 'testing'")))

    def test_patch_success_not_ignore_missing(self):
        self.assertIsNotNone(
            self.db.patch("products/10", ScriptedPatchRequest("this.Name = 'testing'"), ignore_missing=False))

    def test_patch_fail_not_ignore_missing(self):
        with self.assertRaises(DocumentDoesNotExistsException):
            self.db.patch("products/1", ScriptedPatchRequest("this.Name = 'testing'"), ignore_missing=False)


if __name__ == "__main__":
    unittest.main



from pyravendb.connection.requests_factory import HttpRequestsFactory
from pyravendb.custom_exceptions import exceptions
from pyravendb.d_commands import database_commands
from pyravendb.data.document_convention import DocumentConvention
from pyravendb.hilo.hilo_generator import HiloGenerator
from pyravendb.data.database import DatabaseDocument
from pyravendb.store.document_session import documentsession
from pyravendb.tools.utils import Utils
from pyravendb.data.operations import Operations
import traceback
import uuid


class documentstore(object):
    def __init__(self, url=None, database=None, api_key=None):
        self.url = url
        self.database = database
        self.conventions = DocumentConvention()
        self.api_key = api_key
        self._requests_handler = HttpRequestsFactory(url, database, self.conventions, api_key=self.api_key)
        self._database_commands = None
        self._initialize = False
        self.generator = None
        self._operations = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def operations(self):
        self._assert_initialize()
        return self._operations

    @property
    def database_commands(self):
        self._assert_initialize()
        return self._database_commands

    def initialize(self):
        if not self._initialize:
            self._operations = Operations(self._requests_handler)
            self._database_commands = database_commands.DatabaseCommands(self._requests_handler)
            if self.database is None:
                raise exceptions.InvalidOperationException("None database is not valid")
            if not self.database.lower() == self.conventions.system_database:
                path = "Raven/Databases/{0}".format(self.database)
                response = self._requests_handler.check_database_exists("docs?id=" + Utils.quote_key(path))
                # here we unsure database exists if not create new one
                if response.status_code == 404:
                    try:
                        raise exceptions.ErrorResponseException(
                            "Could not open database named: {0}, database does not exists".format(self.database))
                    except exceptions.ErrorResponseException:
                        print(traceback.format_exc())
                        self._database_commands.admin_commands.create_database(
                            DatabaseDocument(self.database, {"Raven/DataDir": "~\\{0}".format(self.database)}))
            self._requests_handler.get_replication_topology()
            self.generator = HiloGenerator(self.conventions.max_ids_to_catch, self._database_commands)
            self._initialize = True

    def _assert_initialize(self):
        if not self._initialize:
            raise exceptions.InvalidOperationException(
                "You cannot open a session or access the database commands before initializing the document store.\
                Did you forget calling initialize()?")

    def open_session(self, database=None, api_key=None, force_read_from_master=False):
        self._assert_initialize()
        session_id = uuid.uuid4()
        database_commands_for_session = self._database_commands
        if database is not None:
            requests_handler = HttpRequestsFactory(self.url, database, self.conventions, force_get_topology=True,
                                                   api_key=api_key)
            path = "Raven/Databases/{0}".format(database)
            response = requests_handler.check_database_exists("docs?id=" + Utils.quote_key(path))
            if response.status_code != 200:
                raise exceptions.ErrorResponseException("Could not open database named:{0}".format(database))
            database_commands_for_session = database_commands.DatabaseCommands(requests_handler)
        return documentsession(database, self, database_commands_for_session, session_id, force_read_from_master)

    def generate_id(self, entity):
        return self.generator.generate_document_id(entity, self.conventions, self._requests_handler)



from pyravendb.custom_exceptions import exceptions
from pyravendb.tools.utils import Utils
from threading import Lock


class Range(object):
    def __init__(self, min_id=1, max_id=0):
        self.min_id = min_id
        self.max_id = max_id
        self.current = self.min_id - 1


class HiloGenerator(object):
    def __init__(self, capacity, database_commands):
        self.capacity = capacity
        self.collection_ranges = {}
        self.lock = Lock()
        self.database_commands = database_commands

    def generate_document_id(self, entity, convention, request_handler):
        """
        @param entity: the object we want to generate id for
        :type object
        @param convention: the document_store convention
        :type Convention
        @param request_handler: the handler for the requests
        :type HttpRequestsFactory
        @return: the id
        :rtype: int
        """
        type_tag_name = convention.default_transform_type_tag_name(entity.__class__.__name__)
        if type_tag_name in self.collection_ranges:
            self.next_id(type_tag_name, request_handler)
        else:
            with self.lock:
                if type_tag_name in self.collection_ranges:
                    self.next_id(type_tag_name, request_handler)
                    return "{0}/{1}".format(type_tag_name, self.collection_ranges[type_tag_name].current)
                ranges = {type_tag_name: Range()}
                ranges.update(self.collection_ranges)
                self.collection_ranges = ranges
            self.next_id(type_tag_name, request_handler)
        return "{0}/{1}".format(type_tag_name, self.collection_ranges[type_tag_name].current)

    def next_id(self, type_tag_name, request_handler):
        while True:
            my_collection_range = self.collection_ranges[type_tag_name]
            current_max = my_collection_range.max_id
            with self.lock:
                my_collection_range.current += 1
                current_id = my_collection_range.current
            if current_id <= current_max:
                return
            with self.lock:
                if my_collection_range != self.collection_ranges[type_tag_name]:
                    continue
                try:
                    self.collection_ranges[type_tag_name] = self.get_next_range(type_tag_name, request_handler)
                except exceptions.FetchConcurrencyException:
                    pass

    def get_next_range(self, type_tag_name, request_handler):
        while True:
            path = "Raven/Hilo/{0}&id=Raven/ServerPrefixForHilo".format(type_tag_name)
            document = None
            try:
                document = self.database_commands.get(path)["Results"][0]
            except (exceptions.ErrorResponseException, IndexError):
                pass
            min_id = 1
            if document is None:
                etag = Utils.empty_etag()
                max_id = self.capacity
            else:
                min_id = document["Max"] + 1
                max_id = document["Max"] + self.capacity
                etag = document["@metadata"]["@etag"]

            try:
                request_handler.call_hilo(type_tag_name, max_id, etag)
                return Range(min_id, max_id)
            except exceptions.FetchConcurrencyException:
                pass



from enum import Enum
from pyravendb.custom_exceptions.exceptions import *
from pyravendb.data.indexes import IndexQuery
from pyravendb.tools.utils import Utils
from datetime import timedelta
import sys
import time
import re


class EscapeQueryOptions(Enum):
    EscapeAll = 0
    AllowPostfixWildcard = 1
    AllowAllWildcards = 2
    RawQuery = 3


class Query(object):
    def __init__(self, session):
        self.session = session
        self.query_builder = ""
        self.negate = False
        self._sort_hints = set()
        self._sort_fields = set()
        self.fetch = None
        self._page_size = None
        self._start = None

    def __call__(self, object_type=None, index_name=None, using_default_operator=None,
                 wait_for_non_stale_results=False, includes=None, with_statistics=False, nested_object_types=None):
        """
        @param index_name: The index name we want to apply
        :type index_name: str
        @param object_type: The type of the object we want to track the entity too
        :type Type
        @param nested_object_types: A dict of classes for nested object the key will be the name of the class and the
        value will be the object we want to get for that attribute
        :type str
        @param using_default_operator: If None, by default will use OR operator for the query (can use for OR or AND)
        @param with_statistics: Make it True to get the query statistics as well
        :type bool
        """
        if not index_name:
            index_name = "dynamic"
            if object_type is not None:
                index_name += "/{0}".format(self.session.conventions.default_transform_plural(object_type.__name__))
        self.index_name = index_name
        self.object_type = object_type
        self.nested_object_types = nested_object_types
        self.using_default_operator = using_default_operator
        self.wait_for_non_stale_results = wait_for_non_stale_results
        self.includes = includes
        if includes and not isinstance(self.includes, list):
            self.includes = [self.includes]
        self._with_statistics = with_statistics
        return self

    def select(self, *args):
        """
        Fetch only the required fields from the server

        @param args: The name of the terms you like to acquire

        """
        if args:
            self.fetch = args
        return self

    def _lucene_builder(self, value, action=None, escape_query_options=EscapeQueryOptions.EscapeAll):

        if isinstance(value, str):
            if escape_query_options == EscapeQueryOptions.EscapeAll:
                value = Utils.escape(value, False, False)

            elif escape_query_options == EscapeQueryOptions.AllowPostfixWildcard:
                value = Utils.escape(value, False, False)
            elif escape_query_options == EscapeQueryOptions.AllowAllWildcards:
                value = Utils.escape(value, True, False)
                value = re.sub(r'"\\\*(\s|$)"', "*${1}", value)
            elif escape_query_options == EscapeQueryOptions.RawQuery:
                value = Utils.escape(value, False, False).replace("\\*", "*")

        lucene_text = Utils.to_lucene(value, action=action)

        if len(self.query_builder) > 0 and not self.query_builder.endswith(' '):
            self.query_builder += ' '
        if self.negate:
            self.negate = False
            self.query_builder += '-'

        return lucene_text

    def __iter__(self):
        return self._execute_query().__iter__()

    def where_equals(self, field_name, value, escape_query_options=EscapeQueryOptions.EscapeAll):
        """
        To get all the document that equal to the value in the given field_name

        @param field_name:The field name in the index you want to query.
        :type str
        @param value: The value will be the fields value you want to query
        @param escape_query_options: the way we should escape special characters
        :type EscapeQueryOptions
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        if value is not None and not isinstance(value, str) and field_name is not None:
            sort_hint = self.session.conventions.get_default_sort_option(type(value).__name__)
            if sort_hint:
                if sys.version_info.major > 2:
                    if value > 2147483647:
                        sort_hint = self.session.conventions.get_default_sort_option("long")
                self._sort_hints.add("SortHint-{0}={1}".format(field_name, sort_hint))

        lucene_text = self._lucene_builder(value, action="equal", escape_query_options=escape_query_options)
        self.query_builder += "{0}:{1}".format(field_name, lucene_text)
        return self

    def where(self, **kwargs):
        """
        To get all the document that equal to the value within kwargs with the specific key

        @param kwargs: the keys of the kwargs will be the fields name in the index you want to query.
        The value will be the the fields value you want to query
        (if kwargs[field_name] is a list it will behave has the where_in method)
        """
        for field_name in kwargs:
            if isinstance(kwargs[field_name], list):
                self.where_in(field_name, kwargs[field_name])
            else:
                self.where_equals(field_name, kwargs[field_name])
        return self

    def search(self, field_name, search_terms, escape_query_options=EscapeQueryOptions.RawQuery):
        """
        for more complex text searching

        @param field_name:The field name in the index you want to query.
        :type str
        @param search_terms: the terms you want to query
        :type str
        @param escape_query_options: the way we should escape special characters
        :type EscapeQueryOptions
        """
        search_terms = Utils.quote_key(str(search_terms))
        search_terms = self._lucene_builder(search_terms, "search", escape_query_options)
        self.query_builder += "{0}:{1}".format(field_name, search_terms)
        return self

    def where_ends_with(self, field_name, value):
        """
        To get all the document that ends with the value in the giving field_name

        @param field_name:The field name in the index you want to query.
        :type str
        @param value: The value will be the fields value you want to query
        :type str
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        if value is not None and not isinstance(value, str) and field_name is not None:
            sort_hint = self.session.conventions.get_default_sort_option(type(value).__name__)
            if sort_hint:
                field_name = "{0}_Range".format(field_name)
                if sys.version_info.major > 2:
                    if value > 2147483647:
                        sort_hint = self.session.conventions.get_default_sort_option("long")
                self._sort_hints.add("SortHint-{0}={1}".format(field_name, sort_hint))

        lucene_text = self._lucene_builder(value, action="end_with")
        self.query_builder += "{0}:*{1}".format(field_name, lucene_text)
        return self

    def where_starts_with(self, field_name, value):
        """
        To get all the document that starts with the value in the giving field_name

        @param field_name:The field name in the index you want to query.
        :type str
        @param value: The value will be the fields value you want to query
        :type str
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        if value is not None and not isinstance(value, str) and field_name is not None:
            sort_hint = self.session.conventions.get_default_sort_option(type(value).__name__)
            if sort_hint:
                field_name = "{0}_Range".format(field_name)
                if sys.version_info.major > 2:
                    if value > 2147483647:
                        sort_hint = self.session.conventions.get_default_sort_option("long")
                self._sort_hints.add("SortHint-{0}={1}".format(field_name, sort_hint))

        lucene_text = self._lucene_builder(value, action="end_with")
        self.query_builder += "{0}:{1}*".format(field_name, lucene_text)
        return self

    def where_in(self, field_name, values):
        """
        Check that the field has one of the specified values

        @param field_name: Name of the field
        :type str
        @param values: The values we wish to query
        :type list
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")
        lucene_text = self._lucene_builder(values, action="in")
        if lucene_text is None:
            self.query_builder += "@emptyIn<{0}>:(no-result)".format(field_name)
        else:
            self.query_builder += "@in<{0}>:{1}".format(field_name, lucene_text)
        return self

    def where_between(self, field_name, start, end):
        if isinstance(start, timedelta):
            start = Utils.timedelta_tick(start)
        if isinstance(end, timedelta):
            end = Utils.timedelta_tick(end)

        value = start or end
        if self.session.conventions.uses_range_type(value) and not field_name.endswith("_Range"):
            sort_hint = self.session.conventions.get_default_sort_option(type(value).__name__)
            if sort_hint:
                field_name = "{0}_Range".format(field_name)
                if sys.version_info.major > 2:
                    if value > 2147483647:
                        sort_hint = self.session.conventions.get_default_sort_option("long")
                self._sort_hints.add("SortHint-{0}={1}".format(field_name, sort_hint))

        lucene_text = self._lucene_builder([start, end], action="between")
        self.query_builder += "{0}:{1}".format(field_name, lucene_text)
        return self

    def where_between_or_equal(self, field_name, start, end):
        if isinstance(start, timedelta):
            start = Utils.timedelta_tick(start)
        if isinstance(end, timedelta):
            end = Utils.timedelta_tick(end)

        value = start or end
        if self.session.conventions.uses_range_type(value) and not field_name.endswith("_Range"):
            value_type = type(value).__name__
            if sys.version_info.major > 2:
                if value > 2147483647:
                    value_type = "long"
            sort_hint = self.session.conventions.get_default_sort_option(value_type)
            if sort_hint:
                field_name = "{0}_Range".format(field_name)
                if sys.version_info.major > 2:
                    if value > 2147483647:
                        sort_hint = self.session.conventions.get_default_sort_option("long")
                self._sort_hints.add("SortHint-{0}={1}".format(field_name, sort_hint))
        lucene_text = self._lucene_builder([start, end], action="equal_between")
        self.query_builder += "{0}:{1}".format(field_name, lucene_text)
        return self

    def take(self, count):
        # This method control on the number of the server return results
        self._page_size = count
        return self

    def skip(self, count):
        # This method tell the server to return the results from count and on
        self._start = count
        return self

    def where_greater_than(self, field_name, value):
        return self.where_between(field_name, value, None)

    def where_greater_than_or_equal(self, field_name, value):
        return self.where_between_or_equal(field_name, value, None)

    def where_less_than(self, field_name, value):
        return self.where_between(field_name, None, value)

    def where_less_than_or_equal(self, field_name, value):
        return self.where_between_or_equal(field_name, None, value)

    def where_not_none(self, field_name):
        if len(self.query_builder) > 0:
            self.query_builder += ' '
        self.query_builder += '('
        self.where_equals(field_name, '*',
                          escape_query_options=EscapeQueryOptions.RawQuery).and_also().add_not().where_equals(
            field_name, None)
        self.query_builder += ')'
        return self

    def order_by(self, fields_name):
        if isinstance(fields_name, list):
            for field_name in fields_name:
                self._sort_fields.add(field_name)
        else:
            self._sort_fields.add(fields_name)
        return self

    def order_by_descending(self, fields_name):
        if isinstance(fields_name, list):
            for i in range(len(fields_name)):
                fields_name[i] = "-{0}".format(fields_name[i])
        else:
            fields_name = "-{0}".format(fields_name)
        self.order_by(fields_name)
        return self

    def and_also(self):
        if len(self.query_builder) > 0:
            self.query_builder += " AND"
        return self

    def or_else(self):
        if len(self.query_builder) > 0:
            self.query_builder += " OR"
        return self

    def add_not(self):
        self.negate = True
        return self

    def boost(self, value):
        if len(self.query_builder) < 1:
            raise InvalidOperationException("Missing where clause")
        if value < 0:
            raise ArgumentOutOfRangeException("boost", "boost factor must be a positive number")
        if value != 1:
            # 1 is the default
            if self.query_builder.endswith(')'):
                self.query_builder = "{0}^{1}".format(self.query_builder, value)
            else:
                self.query_builder += "^{0}".format(value)
        return self

    def get_index_query(self):
        return IndexQuery(self.query_builder, default_operator=self.using_default_operator,
                          sort_hints=self._sort_hints, sort_fields=self._sort_fields,
                          fetch=self.fetch,
                          wait_for_non_stale_results=self.wait_for_non_stale_results,
                          start=self._start)

    def _execute_query(self):
        self.session.increment_requests_count()
        conventions = self.session.conventions
        end_time = time.time() + conventions.timeout
        while True:
            index_query = self.get_index_query()
            if self._page_size is not None:
                index_query.page_size = self._page_size
            response = self.session.database_commands.query(self.index_name, index_query, includes=self.includes)
            if response["IsStale"] and self.wait_for_non_stale_results:
                if time.time() > end_time:
                    raise ErrorResponseException("The index is still stale after reached the timeout")
                    time.sleep(0.1)
                continue
            break

        results = []
        response_results = response.pop("Results")
        response_includes = response.pop("Includes")

        for result in response_results:
            entity, metadata, original_metadata = Utils.convert_to_entity(result, self.object_type, conventions,
                                                                          self.nested_object_types,
                                                                          fetch=False if not self.fetch else True)
            if self.object_type != dict and not self.fetch:
                self.session.save_entity(key=original_metadata.get("@id", None), entity=entity,
                                         original_metadata=original_metadata,
                                         metadata=metadata, document=result)
            results.append(entity)
        self.session.save_includes(response_includes)
        if self._with_statistics:
            return results, response
        return results


class QueryOperator(Enum):
    OR = "OR"
    AND = "AND"



class InvalidOperationException(Exception):
    pass


class ErrorResponseException(Exception):
    pass


class DocumentDoesNotExistsException(Exception):
    pass


class NonUniqueObjectException(Exception):
    pass


class FetchConcurrencyException(Exception):
    pass


class ArgumentOutOfRangeException(Exception):
    pass


class TimeoutException(Exception):
    pass


class NotSupportedException(Exception):
    pass


from pyravendb.connection.requests_factory import HttpRequestsFactory
from pyravendb.data.indexes import IndexDefinition
import unittest
import sys
import os

sys.path.append(os.path.abspath(__file__ + "/../../"))

from pyravendb.d_commands import database_commands
from pyravendb.data.database import DatabaseDocument


class TestBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.default_url = "http://localhost:8080"
        cls.default_database = "NorthWindTest"
        cls.request_handler = HttpRequestsFactory(cls.default_url, cls.default_database)
        cls.db = database_commands.DatabaseCommands(cls.request_handler)
        while True:
            try:
                cls.db.admin_commands.create_database(DatabaseDocument(cls.default_database, {"Raven/DataDir": "test"}))
                break
            except Exception as e:
                if "already exists" in str(e):
                    cls.db.admin_commands.delete_database(cls.default_database, True)

        cls.index_map = ("from doc in docs "
                         "select new{"
                         "Tag = doc[\"@metadata\"][\"Raven-Entity-Name\"],"
                         "LastModified = (DateTime)doc[\"@metadata\"][\"Last-Modified\"],"
                         "LastModifiedTicks = ((DateTime)doc[\"@metadata\"][\"Last-Modified\"]).Ticks}"
                         )
        cls.index = IndexDefinition(cls.index_map)
        cls.db.put_index("Testing", cls.index, True)

    @classmethod
    def tearDownClass(cls):
        cls.db.admin_commands.delete_database("NorthWindTest", True)



import unittest
from pyravendb.tests.test_base import TestBase


class TestDelete(TestBase):
    def test_put_success(self):
        self.db.put("testing/1", {"Name": "test"}, None)
        response = self.db.get("testing/1")
        self.assertEqual(response["Results"][0]["@metadata"]["@id"], "testing/1")

    def test_put_fail(self):
        with self.assertRaises(ValueError):
            self.db.put("testing/2", "document", None)


if __name__ == "__main__":
    unittest.main()



from pyravendb.custom_exceptions import exceptions
import time


class BulkOperationOption(object):
    """
      :param allow_stale Indicates whether operations are allowed on stale indexes.
      :param stale_timeout If AllowStale is set to false and index is stale, then this is the maximum timeout to wait
      for index to become non-stale. If timeout is exceeded then exception is thrown.
      None by default - throw immediately if index is stale.
      :param max_ops_per_sec Limits the amount of base operation per second allowed.
      :param retrieve_details Determines whether operation details about each document should be returned by server.
  """

    def __init__(self, allow_stale=True, stale_timeout=None, max_ops_per_sec=None, retrieve_details=False):
        self.allow_stale = allow_stale
        self.stale_timeout = stale_timeout
        self.retrieve_details = retrieve_details
        self.max_ops_per_sec = max_ops_per_sec


class Operations(object):
    def __init__(self, request_handler):
        self.request_handler = request_handler

    def wait_for_operation_complete(self, operation_id, timeout=None):
        start_time = time.time()
        try:
            path = "operation/status?id={0}".format(operation_id)
            while True:
                response = self.request_handler.http_request_handler(path, "GET")
                if response.status_code == 200:
                    response = response.json()
                if timeout and time.time() - start_time > timeout:
                    raise exceptions.TimeoutException("The operation did not finish before the timeout end")
                if response["Faulted"]:
                    if "Error" in response["State"]:
                        error = response["State"]["Error"]
                    else:
                        error = "Something went wrong with the operation"
                    raise exceptions.InvalidOperationException(error)
                if response["Completed"]:
                    return response
                time.sleep(0.5)
        except ValueError as e:
            raise exceptions.InvalidOperationException(e)



from pyravendb.data.indexes import SortOptions
from datetime import datetime, timedelta
from pyravendb.tools.utils import Utils
from enum import Enum
from inflector import Inflector
import sys


class Failover(Enum):
    """
    Allow to read from the secondary server(s), but immediately fail writes
    to the secondary server(s).

    This is usually the safest approach, because it means that you can still serve
    read requests when the primary node is down, but don't have to deal with replication
    conflicts if there are writes to the secondary when the primary node is down.
    """
    allow_reads_from_secondaries = 1,

    """
    Allow reads from and writes to secondary server(s).
    Choosing this option requires that you'll have some way of propagating changes
    made to the secondary server(s) to the primary node when the primary goes back
    up.
    A typical strategy to handle this is to make sure that the replication is setup
    in a master/master relationship, so any writes to the secondary server will be
    replicated to the master server.
    Please note, however, that this means that your code must be prepared to handle
    conflicts in case of different writes to the same document across nodes.
    """

    allow_reads_from_secondaries_and_writes_to_secondaries = 3,

    """
    Immediately fail the request, without attempting any failover. This is true for both
    reads and writes. The RavenDB client will not even check that you are using replication.
    This is mostly useful when your replication setup is meant to be used for backups / external
    needs, and is not meant to be a failover storage.
    """

    fail_immediately = 0,

    """
    Read requests will be spread across all the servers, instead of doing all the work against the master.
    Write requests will always go to the master.
    This is useful for striping, spreading the read load among multiple servers. The idea is that this will give us
    better read performance overall.
    A single session will always use the same server, we don't do read striping within a single session.
    Note that using this means that you cannot set UserOptimisticConcurrency to true,
    because that would generate concurrency exceptions.
    If you want to use that, you have to open the session with ForceReadFromMaster set to true.
    """
    read_from_all_servers = 1024


inflector = Inflector()


class DocumentConvention(object):
    def __init__(self):
        self.max_number_of_request_per_session = 30
        self.max_ids_to_catch = 32
        # timeout for wait to server in seconds
        self.timeout = 30
        self.failover_behavior = Failover.allow_reads_from_secondaries
        self.default_use_optimistic_concurrency = False
        self.json_default_method = DocumentConvention.json_default
        self._system_database = "system"

    @staticmethod
    def json_default(o):
        if isinstance(o, datetime):
            return Utils.datetime_to_string(o)
        elif isinstance(o, timedelta):
            return Utils.timedelta_to_str(o)
        elif getattr(o, "__dict__", None):
            return o.__dict__
        else:
            raise TypeError(repr(o) + " is not JSON serializable (Try add a json default method to store convention)")

    @staticmethod
    def default_transform_plural(name):
        return inflector.conditional_plural(2, name)

    @staticmethod
    def default_transform_type_tag_name(name):
        count = sum(1 for c in name if c.isupper())
        if count <= 1:
            return DocumentConvention.default_transform_plural(name.lower())
        return DocumentConvention.default_transform_plural(name)

    @staticmethod
    def build_default_metadata(entity):
        if entity is None:
            return {}
        return {"Raven-Entity-Name": DocumentConvention.default_transform_plural(entity.__class__.__name__),
                "Raven-Python-Type": "{0}.{1}".format(entity.__class__.__module__, entity.__class__.__name__)}

    @staticmethod
    def try_get_type_from_metadata(metadata):
        if "Raven-Python-Type" in metadata:
            return metadata["Raven-Python-Type"]
        return None

    @staticmethod
    def uses_range_type(obj):
        if obj is None:
            return False
        if sys.version_info.major > 2:
            if isinstance(obj, int) or isinstance(obj, float):
                return True
        else:
            if isinstance(obj, int) or isinstance(obj, float) or isinstance(obj, long):
                return True
        return False

    @staticmethod
    def get_default_sort_option(type_name):
        if not type_name:
            return None
        if type_name == "int" or type_name == "float":
            try:
                return SortOptions.float.value
            except AttributeError:
                return SortOptions.float
        if type_name == "long":
            try:
                return SortOptions.long.value
            except AttributeError:
                return SortOptions.long

    @property
    def system_database(self):
        return self._system_database



import sys
import os

sys.path.append(os.path.abspath(__file__ + "/../"))

from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import documentstore
from pyravendb.custom_exceptions import exceptions
import unittest


class TestDelete(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestDelete, cls).setUpClass()
        cls.db.put("products/101", {"name": "test"}, {"Raven-Python-Type": "delete_test.Product"})
        cls.db.put("products/10", {"name": "test"}, {})
        cls.db.put("products/106", {"name": "test"}, {})
        cls.db.put("products/107", {"name": "test"}, {})
        cls.document_store = documentstore(cls.default_url, cls.default_database)
        cls.document_store.initialize()

    def test_delete_with_key_with_save_session(self):
        with self.document_store.open_session() as session:
            session.delete("products/101")
            session.save_changes()
            self.assertIsNone(session.load("products/101"))

    def test_delete_with_key_without_save_session(self):
        with self.document_store.open_session() as session:
            session.delete("products/10")
            self.assertIsNone(session.load("products/101"))

    def test_delete_after_change_fail(self):
        with self.document_store.open_session() as session:
            product = session.load("products/106")
            product.name = "testing"
            with self.assertRaises(exceptions.InvalidOperationException):
                session.delete("products/106")

    def test_delete_after_change_success_with_save_session(self):
        with self.document_store.open_session() as session:
            product = session.load("products/107")
            product.name = "testing"
            session.delete_by_entity(product)
            session.save_changes()
            self.assertIsNone(session.load("products/107"))


if __name__ == "__main__":
    unittest.main()



import ijson
import re

LEXEME_RE = re.compile(r'[a-z0-9eE\.\+-]+|\S')


# The code imported from ijson to be able to receive json from socket
class IncrementalJsonParser:
    def __init__(self, socket):
        self.lexer = self.lexer(socket)

    @staticmethod
    def basic_parse(response, buf_size=ijson.backend.BUFSIZE):
        '''
        Iterator yielding unprefixed events.

        Parameters:

        - response: a stream response from requests
        '''
        lexer = iter(IncrementalJsonParser.lexer(response, buf_size))
        for value in ijson.backend.parse_value(lexer):
            yield value
        try:
            next(lexer)
        except StopIteration:
            pass
        else:
            raise ijson.common.JSONError('Additional data')

    @staticmethod
    def lexer(response, buf_size=ijson.backend.BUFSIZE):
        it = response.iter_content(buf_size)
        data = next(it, None)
        if data is None:
            yield None
        buf = data.decode('utf-8')
        pos = 0
        discarded = 0
        while True:
            match = LEXEME_RE.search(buf, pos)
            if pos < len(buf) and match:
                lexeme = match.group()
                if lexeme == '"':
                    pos = match.start()
                    start = pos + 1
                    while True:
                        try:
                            end = buf.index('"', start)
                            escpos = end - 1
                            while buf[escpos] == '\\':
                                escpos -= 1
                            if (end - escpos) % 2 == 0:
                                start = end + 1
                            else:
                                break
                        except ValueError:
                            data = next(it, buf_size)
                            if not data:
                                yield None
                            buf += data.decode('utf-8')
                    yield discarded + pos, buf[pos:end + 1]
                    pos = end + 1
                else:
                    while match.end() == len(buf):
                        data = next(data, None)
                        if not data:
                            break
                        buf += data
                        match = LEXEME_RE.search(buf, pos)
                        lexeme = match.group()
                    pos = match.end()
                    yield discarded + match.start(), lexeme
            else:
                data = next(it, None)
                if data is None:
                    yield None
                buf = data.decode('utf-8')
                pos = 0

    # def create_array(self, gen):
    #     arr = []
    #
    #     for (token, val) in gen:
    #         if token == "end_array":
    #             return arr
    #         arr.append(self.get_value(gen))
    #
    #     raise ParseError("End array expected, but the generator ended before we got it")
    #
    # def get_value(self, gen):
    #     (token, val) = next(gen)
    #
    #     if token == "start_array":
    #         return self.create_array(gen)
    #
    #     if token == "start_map":
    #         return self.create_object(gen)
    #
    #     return val
    #
    # def create_object(self, gen):
    #     obj = {}
    #
    #     for (token, val) in gen:
    #         if token == "end_map":
    #             return obj
    #         if token == "map_key":
    #             obj[val] = self.get_value(gen)
    #
    #     raise ParseError("End object expected, but the generator ended before we got it")
    #
    # def next_object(self):
    #
    #     try:
    #         (_, text) = next(self.lexer)
    #     except StopIteration:
    #         return None
    #
    #     if text != '{':
    #         raise ParseError("Expected start object, got: " + text)
    #
    #     gen = IncrementalJsonParser.parse_object(self.lexer)
    #     (token, val) = next(gen)
    #     assert token == "start_map"
    #
    #     return self.create_object(gen)
    #
    # @staticmethod
    # def parse_value(lexer, symbol=None, pos=0):
    #     try:
    #         if symbol is None:
    #             pos, symbol = next(lexer)
    #         if symbol == 'null':
    #             yield ('null', None)
    #         elif symbol == 'true':
    #             yield ('boolean', True)
    #         elif symbol == 'false':
    #             yield ('boolean', False)
    #         elif symbol == '[':
    #             for event in IncrementalJsonParser.parse_array(lexer):
    #                 yield event
    #         elif symbol == '{':
    #             for event in IncrementalJsonParser.parse_object(lexer):
    #                 yield event
    #         elif symbol[0] == '"':
    #             yield ('string', IncrementalJsonParser.unescape(symbol[1:-1]))
    #         else:
    #             # if we got a partial token for false / null / true we need to read from the network again
    #             while symbol[0] in ('t', 'n') and len(symbol) < 4 or symbol[0] == 'f' and len(symbol) < 5:
    #                 _, nextpart = next(lexer)
    #                 if symbol == 'null':
    #                     yield ('null', None)
    #                 elif symbol == 'true':
    #                     yield ('boolean', True)
    #                 elif symbol == 'false':
    #                     yield ('boolean', False)
    #                 return
    #
    #             try:
    #                 yield ('number', symbol)
    #             except Exception as e:
    #                 print(e)
    #                 raise ijson.basic_parse.UnexpectedSymbol(symbol, pos)
    #     except StopIteration:
    #         raise ijson.basic_parse.common.IncompleteJSONError('Incomplete JSON data')
    #
    # @staticmethod
    # def parse_array(lexer):
    #     yield ('start_array', None)
    #     try:
    #         pos, symbol = next(lexer)
    #         if symbol != ']':
    #             while True:
    #                 for event in IncrementalJsonParser.parse_value(lexer, symbol, pos):
    #                     yield event
    #                 pos, symbol = next(lexer)
    #                 if symbol == ']':
    #                     break
    #                 if symbol != ',':
    #                     raise ijson.basic_parse.UnexpectedSymbol(symbol, pos)
    #                 pos, symbol = next(lexer)
    #         yield ('end_array', None)
    #     except StopIteration:
    #         raise ijson.basic_parse.common.IncompleteJSONError('Incomplete JSON data')
    #
    # @staticmethod
    # def parse_object(lexer):
    #     yield ('start_map', None)
    #     try:
    #         pos, symbol = next(lexer)
    #         if symbol != '}':
    #             while True:
    #                 if symbol[0] != '"':
    #                     raise ijson.basic_parse.UnexpectedSymbol(symbol, pos)
    #                 yield ('map_key', IncrementalJsonParser.unescape(symbol[1:-1]))
    #                 pos, symbol = next(lexer)
    #                 if symbol != ':':
    #                     raise ijson.basic_parse.UnexpectedSymbol(symbol, pos)
    #                 for event in IncrementalJsonParser.parse_value(lexer, None, pos):
    #                     yield event
    #                 pos, symbol = next(lexer)
    #                 if symbol == '}':
    #                     break
    #                 if symbol != ',':
    #                     raise ijson.basic_parse.UnexpectedSymbol(symbol, pos)
    #                 pos, symbol = next(lexer)
    #         yield ('end_map', None)
    #     except StopIteration:
    #         raise ijson.basic_parse.common.IncompleteJSONError('Incomplete JSON data')
    #
    # @staticmethod
    # def unescape(s):
    #     start = 0
    #     result = ''
    #     while start < len(s):
    #         pos = s.find('\\', start)
    #         if pos == -1:
    #             if start == 0:
    #                 return s
    #             result += s[start:]
    #             break
    #         result += s[start:pos]
    #         pos += 1
    #         esc = s[pos]
    #         if esc == 'u':
    #             result += chr(int(s[pos + 1:pos + 5], 16))
    #             pos += 4
    #         elif esc == 'b':
    #             result += '\b'
    #         elif esc == 'f':
    #             result += '\f'
    #         elif esc == 'n':
    #             result += '\n'
    #         elif esc == 'r':
    #             result += '\r'
    #         elif esc == 't':
    #             result += '\t'
    #         else:
    #             result += esc
    #         start = pos + 1
    #     return result



from enum import Enum


class IndexLockMode(Enum):
    Unlock = 0
    LockedIgnore = 1
    LockedError = 2
    SideBySide = 3

    def __str__(self):
        return self.name


# For server
class SortOptions(Enum):
    # Sort using term values as encoded Floats.  Sort values are Float and
    # lower values are at the front.
    long = "Long"
    # Sort using term values as encoded Doubles.
    # Sort values are Double and lower values are at the front.
    float = "Double"
    custom = "Custom"

    def __str__(self):
        return self.value


class FieldIndexing(Enum):
    # Do not index the field value.
    no = "No"
    # Index the tokens produced by running the field's value through an Analyzer
    analyzed = "Analyzed"
    # Index the field's value without using an Analyzer, so it can be searched.
    not_analyzed = "NotAnalyzed"

    def __str__(self):
        return self.value


class IndexDefinition(object):
    def __init__(self, index_map, name=None, **kwargs):
        """
        @param index_map: The map of the index
        :type str or tuple
        @param name: The name of the index
        :type str
        @param kwargs: Can be use to initialize the other option in the index definition
        :type kwargs
        """
        self.name = name
        self.reduce = kwargs.get("reduce", None)
        self.analyzers = kwargs.get("analyzers", {})
        self.disable_in_memory_indexing = kwargs.get("disable_in_memory_indexing", False)
        self.fields = kwargs.get("fields", [])
        self.index_id = kwargs.get("index_id", 0)
        self.indexes = kwargs.get("indexes", {})
        self.internal_fields_mapping = kwargs.get("internal_fields_mapping", {})
        self._is_compiled = False
        self.is_side_by_side_index = kwargs.get("is_side_by_side_index", False)
        self.is_test_index = kwargs.get("is_test_index", False)
        self.lock_mod = kwargs.get("lock_mod", IndexLockMode.Unlock)
        self.max_index_outputs_per_document = kwargs.get("max_index_outputs_per_document", None)
        self.sort_options = kwargs.get("sort_options", {})
        self.spatial_indexes = kwargs.get("spatial_indexes", {})
        self.stores = kwargs.get("stores", {})
        self.suggestions = kwargs.get("suggestions", {})
        self.term_vectors = kwargs.get("term_vectors", {})
        self.maps = (index_map,) if isinstance(index_map, str) else tuple(set(index_map, ))

    @property
    def type(self):
        if self.name and self.name.startswith('Auto/'):
            return "Auto"
        if self._is_compiled:
            return "Compiled"
        if self.is_map_reduce:
            return "MapReduce"
        return "Map"

    @property
    def is_map_reduce(self):
        return True if self.reduce else False

    @property
    def map(self):
        if not isinstance(self.maps, str):
            return self.maps[0]
        return self.maps

    @map.setter
    def map(self, value):
        if len(self.maps) != 0:
            self.maps.pop()
        self.maps.add(value)

    def to_json(self):
        return {"Analyzers": self.analyzers, "DisableInMemoryIndexing": self.disable_in_memory_indexing,
                "Fields": self.fields, "Indexes": {key: str(self.indexes[key]) for key in self.indexes},
                "IndexId": self.index_id,
                "InternalFieldsMapping": self.internal_fields_mapping, "IsCompiled": self._is_compiled,
                "IsMapReduce": self.is_map_reduce, "IsSideBySideIndex": self.is_side_by_side_index,
                "IsTestIndex": self.is_test_index, "LockMode": str(self.lock_mod), "Map": self.map,
                "Maps": self.maps,
                "MaxIndexOutputsPerDocument": self.max_index_outputs_per_document, "Name": self.name,
                "Reduce": self.reduce, "SortOptions": {key: str(self.sort_options[key]) for key in self.sort_options},
                "SpatialIndexes": self.spatial_indexes,
                "Stores": self.stores, "Suggestions": self.suggestions, "TermVectors": self.term_vectors,
                "Type": self.type}


class IndexQuery(object):
    def __init__(self, query="", default_operator=None, start=None, **kwargs):
        """
        @param query: Actual query that will be performed (Lucene syntax).
        :type str
        @param default_operator: The operator of the query (AND or OR) the default value is OR
        :type Enum.QueryOperator
        @param start : offset used to skip a number of results from a query
        :type int
        @param page_size : the number of the results in the query response
        :type int
        @param fetch : fetch only the terms you want from the index
        :type list
        """
        self.query = query
        self.__page_size_set = False
        self._page_size = 128
        if "page_size" in kwargs:
            self.page_size = kwargs["page_size"]
        self.default_operator = default_operator
        self.sort_hints = kwargs.get("sort_hints", {})
        self.sort_fields = kwargs.get("sort_fields", {})
        self.fetch = kwargs.get("fetch", [])
        self.wait_for_non_stale_results = kwargs.get("wait_for_non_stale_results", False)
        self.start = start

    @property
    def page_size(self):
        return self._page_size

    @page_size.setter
    def page_size(self, value):
        self._page_size = value
        self.__page_size_set = True



from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexQuery
from pyravendb.tests.test_base import TestBase
import time


class TestQuery(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestQuery, cls).setUpClass()
        cls.db.put("products/10", {"Name": "test"}, {"Raven-Entity-Name": "Products", "Raven-Python-Type": "object"})
        cls.db.put("products/11", {"Name": "test2"}, {"Raven-Entity-Name": "Products", "Raven-Python-Type": "object"})
        cls.db.put("products/12", {"Name": "test3"}, {"Raven-Entity-Name": "Products", "Raven-Python-Type": "object"})

    def test_only_query(self):
        self.db.put_index("Testing", self.index, True)
        response = self.db.query("Testing", IndexQuery("Tag:Products"))
        self.assertEqual(response["Results"][0]["Name"], "test")

    def test_get_only_metadata(self):
        self.db.put_index("Testing", self.index, True)
        response = self.db.query("Testing", IndexQuery("Tag:Products"), metadata_only=True)
        self.assertFalse("Name" in response["Results"][0])

    def test_get_only_index_entries(self):
        self.db.put_index("Testing", self.index, True)
        response = self.db.query("Testing", IndexQuery("Tag:Products"), index_entries_only=True)
        self.assertFalse("@metadata" in response["Results"][0])

    def test_fail(self):
        with self.assertRaises(ValueError):
            self.db.query(None, IndexQuery("Tag:Products"))

    def test_fail_none_response(self):
        with self.assertRaises(exceptions.ErrorResponseException):
            self.db.query("IndexIsNotExists", IndexQuery("Tag:Products"))

    def test_only_query_change_page_size_and_add_start(self):
        self.db.put_index("Testing", self.index, True)
        start = time.time()
        while True and time.time() - start < 30:
            response = self.db.query("Testing",
                                     IndexQuery("Tag:Products", page_size=1, start=1))
            if not response["IsStale"]:
                break
        self.assertEqual(response["Results"][0]["Name"], "test2")



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import documentstore
from pyravendb.custom_exceptions import exceptions
import unittest


class Product(object):
    def __init__(self, name):
        self.name = name


class Company(object):
    def __init__(self, name, product):
        self.name = name
        self.product = product


class Foo(object):
    def __init__(self, name, key):
        self.name = name
        self.key = key


class Time(object):
    def __init__(self, td):
        self.td = td


class TestLoad(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestLoad, cls).setUpClass()
        cls.db.put("products/101", {"name": "test"},
                   {"Raven-Python-Type": Product.__module__+".Product"})
        cls.db.put("products/10", {"name": "test"}, {})
        cls.db.put("orders/105", {"name": "testing_order", "key": 92, "product": "products/101"},
                   {"Raven-Entity-Name": "Orders"})
        cls.db.put("company/1", {"name": "test", "product": {"name": "testing_nested"}}, {})
        cls.document_store = documentstore(cls.default_url, cls.default_database)
        cls.document_store.initialize()

    def test_load_success(self):
        with self.document_store.open_session() as session:
            product = session.load("products/101")
            self.assertEqual(product.name, "test")

    def test_load_missing_document(self):
        with self.document_store.open_session() as session:
            self.assertEqual(session.load("products/0"), None)

    def test_multi_load(self):
        with self.document_store.open_session() as session:
            products = session.load(["products/101", "products/10"])
            self.assertEqual(len(products), 2)

    def test_multi_load_with_duplicate_id(self):
        with self.document_store.open_session() as session:
            products = session.load(["products/101", "products/101", "products/10"])
            self.assertEqual(len(products), 3)
            for product in products:
                self.assertIsNotNone(product)

    def test_load_track_entity(self):
        with self.document_store.open_session() as session:
            product = session.load("products/101")
            self.assertTrue(isinstance(product, Product))

    def test_load_track_entity_with_object_type(self):
        with self.document_store.open_session() as session:
            product = session.load("products/101", object_type=Product)
            self.assertTrue(isinstance(product, Product))

    def test_load_track_entity_with_object_type_and_nested_object(self):
        with self.document_store.open_session() as session:
            company = session.load("company/1", object_type=Company, nested_object_types={"product": Product})
            self.assertTrue(isinstance(company, Company) and isinstance(company.product, Product))

    def test_load_track_entity_with_object_type_fail(self):
        with self.document_store.open_session() as session:
            with self.assertRaises(exceptions.InvalidOperationException):
                session.load("products/101", object_type=Foo)

    def test_load_with_include(self):
        with self.document_store.open_session() as session:
            session.load("orders/105", includes="product")
            session.load("products/101")
        self.assertEqual(session.number_of_requests_in_session, 1)


if __name__ == "__main__":
    unittest.main()



import unittest

from pyravendb.custom_exceptions import exceptions
from pyravendb.tests.test_base import TestBase


class TestDelete(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestDelete, cls).setUpClass()
        cls.response = cls.db.put("products/101", {"Name": "test"}, {})
        cls.db.put("products/10", {"Name": "test"}, {})
        cls.other_response = cls.db.put("products/102", {"Name": "test"}, {})

    def test_delete_success_no_etag(self):
        self.assertIsNone(self.db.delete("products/10"))

    def test_delete_fail(self):
        with self.assertRaises(exceptions.ErrorResponseException):
            self.db.delete("products/101", self.response[0]["Etag"][0:len(self.response[0]["Etag"]) - 2] + "10")

    def test_delete_fail_with_etag(self):
        self.assertIsNone(self.db.delete("products/102", self.other_response[0]["Etag"]))

    if __name__ == "__main__":
        unittest.main()



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import documentstore
from pyravendb.custom_exceptions import exceptions
import unittest


class Foo(object):
    def __init__(self, name, key):
        self.name = name
        self.key = key
        self.Id = None


class TestSessionStore(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestSessionStore, cls).setUpClass()
        cls.document_store = documentstore(cls.default_url, cls.default_database)
        cls.document_store.initialize()

    def test_store_without_key(self):
        foo = Foo("test", 10)
        with self.document_store.open_session() as session:
            session.store(foo)
            session.save_changes()

        with self.document_store.open_session() as session:
            self.assertIsNotNone(session.load(foo.Id))

    def test_store_with_key(self):
        foo = Foo("test", 20)
        with self.document_store.open_session() as session:
            session.store(foo, "testingStore/1")
            session.save_changes()

        with self.document_store.open_session() as session:
            self.assertEqual(session.load("testingStore/1").key, 20)

    def test_store_after_delete_fail(self):
        foo = Foo("test", 20)
        with self.document_store.open_session() as session:
            session.store(foo, "testingStore")
            session.save_changes()

        with self.document_store.open_session() as session:
            session.delete("testingStore")
            with self.assertRaises(exceptions.InvalidOperationException):
                session.store(foo)

    def test_store_fifty_documents_at_once(self):
        with self.document_store.open_session() as session:
            for i in range(40):
                session.store(Foo("Idan", 29))
            session.save_changes()

            result = list(session.query(Foo).where_equals("name", "Idan"))
            self.assertTrue(len(result) >= 40)

    def _save_documents(self, documents):
        with self.document_store.open_session() as session:
            for document in documents:
                session.store(document)
            session.save_changes()

    def test_store_the_same_documents_should_work(self):
        documents = []
        for i in range(0, 40):
            documents.append(Foo("IdanHaim", i))

        self._save_documents(documents)
        self._save_documents(documents)
        self._save_documents(documents)
        self._save_documents(documents)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='pyravendb',
    packages=find_packages(),
    version='3.5.4.1',
    description='This is the official python client for RavenDB document database',
    author='Idan Haim Shalom',
    author_email='haimdude@gmail.com',
    url='https://github.com/IdanHaim/RavenDB-Python-Client',
    license='MIT',
    keywords='pyravendb',
    install_requires=
    [
        "pycrypto >= 2.6.1",
        "requests >= 2.9.1",
        "inflector >= 2.0.11",
        'ijson',
    ],
    zip_safe=False
)



import unittest

from pyravendb.custom_exceptions import exceptions
from pyravendb.tests.test_base import TestBase


class TestIndexActions(TestBase):
    @classmethod
    def setUpClass(cls):
        super(TestIndexActions, cls).setUpClass()

    def test_put_index_success(self):
        assert self.db.put_index("region", self.index, True)

    def test_put_index_fail(self):
        self.db.put_index("test", self.index, True)
        with self.assertRaises(exceptions.InvalidOperationException):
            self.db.put_index("test", self.index, False)

    def test_get_index_success(self):
        self.db.put_index("test", self.index, True)
        self.assertIsNotNone(self.db.get_index("test"))

    def test_get_index_fail(self):
        self.assertIsNone(self.db.get_index("reg"))

    def test_delete_index_success(self):
        self.db.put_index("test", self.index, True)
        self.assertIsNone(self.db.delete_index("delete"))

    def test_delete_index_fail(self):
        with self.assertRaises(ValueError):
            self.db.delete_index(None)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*- #
from pyravendb.data.indexes import IndexQuery
from pyravendb.store.session_query import QueryOperator
from pyravendb.data.patches import ScriptedPatchRequest
from pyravendb.data.indexes import IndexDefinition
from pyravendb.custom_exceptions import exceptions
from pyravendb.tools.utils import Utils
import collections


class DatabaseCommands(object):
    def __init__(self, request_handler):
        self._requests_handler = request_handler
        self.admin_commands = self.Admin(self._requests_handler)

    def change_database(self, database):
        self._requests_handler.database = database
        self._requests_handler.copy = True

    @staticmethod
    def run_async(func, func_parameter=()):

        """
        @param func: The instance of the function you want to run
        :type instancemethod
        @param func_parameter: The function parameter
        :type tuple
        """

        from multiprocessing.pool import ThreadPool
        pool = ThreadPool(processes=1)
        async_result = pool.apply_async(func, func_parameter)
        return async_result.get()

    def get(self, key_or_keys, includes=None, metadata_only=False, force_read_from_master=False):
        """
        @param key_or_keys: the key of the documents you want to retrieve (key can be a list of ids)
        :type str or list
        @param includes: array of paths in documents in which server should look for a 'referenced' document
        :type list
        @param metadata_only: specifies if only document metadata should be returned
        :type bool
        @return: A list of the id or ids we looked for (if they exists)
        :rtype: dict
        @param force_read_from_master: If True the reading also will be from the master
        :type bool
        """
        if key_or_keys is None:
            raise ValueError("None Key is not valid")
        path = "queries/?"
        method = "GET"
        data = None
        if includes:
            path += "".join("&include=" + Utils.quote_key(item) for item in includes)
        # make get method handle a multi document requests in a single request
        if isinstance(key_or_keys, list):
            key_or_keys = collections.OrderedDict.fromkeys(key_or_keys)
            if metadata_only:
                path += "&metadata-only=True"

            # If it is too big, we drop to POST (note that means that we can't use the HTTP cache any longer)
            if (sum(len(x) for x in key_or_keys)) > 1024:
                method = "POST"
                data = list(key_or_keys)
            else:
                path += "".join("&id=" + Utils.quote_key(item) for item in key_or_keys)

        else:
            path += "&id={0}".format(Utils.quote_key(key_or_keys))

        response = self._requests_handler.http_request_handler(path, method, data=data,
                                                               force_read_from_master=force_read_from_master)
        if response.status_code == 200:
            response = response.json()
        return response

    def delete(self, key, etag=None):
        if key is None:
            raise ValueError("None Key is not valid")
        if not isinstance(key, str):
            raise ValueError("key must be {0}".format(type("")))
        headers = {}
        if etag is not None:
            headers["If-None-Match"] = "\"" + etag + "\""
        key = Utils.quote_key(key)
        path = "docs/{0}".format(key)
        response = self._requests_handler.http_request_handler(path, "DELETE", headers=headers)
        if response.status_code != 204:
            raise exceptions.ErrorResponseException(response.json()["Error"])

    def put(self, key, document, metadata=None, etag=None):
        """
        @param key: unique key under which document will be stored
        :type str
        @param document: document data
        :type dict
        @param metadata: document metadata
        :type dict
        @param etag: current document etag, used for concurrency checks (null to skip check)
        :type str
        @return: json file
        :rtype: dict
        """
        headers = None
        if document is None:
            document = {}
        if metadata is None:
            metadata = {}
        if not isinstance(key, str):
            raise ValueError("key must be {0}".format(type("")))
        if not isinstance(document, dict) or not isinstance(metadata, dict):
            raise ValueError("document and metadata must be dict")
        data = [{"Key": key, "Document": document, "Metadata": metadata, "AdditionalData": None,
                 "Method": "PUT", "Etag": etag}]
        if etag:
            headers = {"if-None-Match": "\"" + etag + "\""}
        response = self._requests_handler.http_request_handler("bulk_docs", "POST", data=data, headers=headers).json()
        if "Error" in response:
            if "ActualEtag" in response:
                raise exceptions.FetchConcurrencyException(response["Error"])
            raise exceptions.ErrorResponseException(response["Error"][:85])
        return response

    def batch(self, commands_array):
        data = []
        for command in commands_array:
            if not hasattr(command, 'command'):
                raise ValueError("Not a valid command")
            data.append(command.to_json())
        response = self._requests_handler.http_request_handler("bulk_docs", "POST", data=data).json()
        if "Error" in response:
            raise ValueError(response["Error"])
        return response

    def put_index(self, index_name, index_def, overwrite=False):
        """
        @param index_name:The name of the index
        @param index_def: IndexDefinition class a definition of a RavenIndex
        @param overwrite: if set to True overwrite
        """
        if index_name is None:
            raise ValueError("None index_name is not valid")
        if not isinstance(index_def, IndexDefinition):
            raise ValueError("index_def must be IndexDefinition type")
        index_name = Utils.quote_key(index_name)
        path = "indexes/{0}?definition=yes".format(index_name)
        response = self._requests_handler.http_request_handler(path, "GET")
        if not overwrite and response.status_code != 404:
            raise exceptions.InvalidOperationException("Cannot put index:{0},index already exists".format(index_name))

        data = index_def.to_json()
        return self._requests_handler.http_request_handler(path, "PUT", data=data).json()

    def get_index(self, index_name, force_read_from_master=False):
        """
        @param index_name: Name of the index you like to get or delete
        :type str
        @return: json or None
        :rtype: dict
        @param force_read_from_master: If True the reading also will be from the master
        :type bool
        """
        path = "indexes/{0}?definition=yes".format(Utils.quote_key(index_name))
        response = self._requests_handler.http_request_handler(path, "GET",
                                                               force_read_from_master=force_read_from_master)
        if response.status_code != 200:
            return None
        return response.json()

    def delete_index(self, index_name):
        """
        @param index_name: Name of the index you like to get or delete
        :type str
        @return: json or None
        :rtype: dict
        """
        if not index_name:
            raise ValueError("None or empty index_name is invalid")
        path = "indexes/{0}".format(Utils.quote_key(index_name))
        self._requests_handler.http_request_handler(path, "DELETE")

    def update_by_index(self, index_name, query, scripted_patch=None, options=None):
        """
        @param index_name: name of an index to perform a query on
        :type str
        @param query: query that will be performed
        :type IndexQuery
        @param options: various operation options e.g. AllowStale or MaxOpsPerSec
        :type BulkOperationOptions
        @param scripted_patch: JavaScript patch that will be executed on query results( Used only when update)
        :type ScriptedPatchRequest
        @return: json
        :rtype: dict
        """
        if not isinstance(query, IndexQuery):
            raise ValueError("query must be IndexQuery Type")
        path = Utils.build_path(index_name, query, options, with_page_size=False)
        if scripted_patch:
            if not isinstance(scripted_patch, ScriptedPatchRequest):
                raise ValueError("scripted_patch must be ScriptedPatchRequest Type")
            scripted_patch = scripted_patch.to_json()

        response = self._requests_handler.http_request_handler(path, "EVAL", data=scripted_patch)
        if response.status_code != 200 and response.status_code != 202:
            raise response.raise_for_status()
        return response.json()

    def delete_by_index(self, index_name, query, options=None):
        """
        @param index_name: name of an index to perform a query on
        :type str
        @param query: query that will be performed
        :type IndexQuery
        @param options: various operation options e.g. AllowStale or MaxOpsPerSec
        :type BulkOperationOptions
        @return: json
        :rtype: dict
        """
        path = Utils.build_path(index_name, query, options, with_page_size=False)
        response = self._requests_handler.http_request_handler(path, "DELETE")
        if response.status_code != 200 and response.status_code != 202:
            try:
                raise exceptions.ErrorResponseException(response.json()["Error"][:100])
            except ValueError:
                raise response.raise_for_status()
        return response.json()

    def patch(self, key, scripted_patch, etag=None, ignore_missing=True, default_metadata=None, patch_default=None):
        from pyravendb.d_commands import commands_data
        if default_metadata is None:
            default_metadata = {}
        batch_result = self.batch([
            commands_data.ScriptedPatchCommandData(key, scripted_patch, etag, default_metadata, patch_default)])
        if not ignore_missing and batch_result[0]["PatchResult"] == "DocumentDoesNotExists" and patch_default is None:
            raise exceptions.DocumentDoesNotExistsException("Document with key {0} does not exist.".format(key))
        return batch_result

    def get_operation_status(self, operation_id):
        pass

    @staticmethod
    def _build_query_request_path(path, index_query, includes=None, metadata_only=False, index_entries_only=False,
                                  include_query=True, add_page_size=True):
        if index_query.default_operator is QueryOperator.AND:
            try:
                operator = index_query.default_operator.value
            except AttributeError:
                operator = index_query.default_operator
            path += "&operator={0}".format(operator)
        if index_query.query and include_query:
            path += "&query={0}".format(Utils.quote_key(index_query.query, safe='/'))
        if index_query.sort_hints:
            for hint in index_query.sort_hints:
                path += "&{0}".format(hint)
        if index_query.sort_fields:
            for field in index_query.sort_fields:
                path += "&sort={0}".format(field)
        if index_query.fetch:
            for item in index_query.fetch:
                path += "&fetch={0}".format(item)
        if metadata_only:
            path += "&metadata-only=true"
        if index_entries_only:
            path += "&debug=entries"
        if includes and len(includes) > 0:
            path += "".join("&include=" + item for item in includes)
        if index_query.start and index_query.start != 0:
            path += "&start={0}".format(index_query.start)

        if add_page_size and index_query.page_size:
            path += "&pageSize={0}".format(index_query.page_size)

        if index_query.wait_for_non_stale_results:
            path += "&waitForNonStaleResultsAsOfNow=true"

        return path

    def query(self, index_name, index_query, includes=None, metadata_only=False, index_entries_only=False,
              force_read_from_master=False):
        """
        @param index_name: A name of an index to query
        @param force_read_from_master: If True the reading also will be from the master
        :type bool
        :type str
        @param index_query: A query definition containing all information required to query a specified index.
        :type IndexQuery
        @param includes: An array of relative paths that specify related documents ids
        which should be included in a query result.
        :type list
        @param metadata_only: True if returned documents should include only metadata without a document body.
        :type bool
        @param index_entries_only: True if query results should contain only index entries.
        :type bool
        @return:json
        :rtype:dict
        """
        if not index_name:
            raise ValueError("index_name cannot be None or empty")
        if index_query is None:
            raise ValueError("None query is invalid")
        if not isinstance(index_query, IndexQuery):
            raise ValueError("index_query must be IndexQuery type")
        path = "indexes/{0}?".format(Utils.quote_key(index_name))
        path += self._build_query_request_path(path=path, index_query=index_query, includes=includes,
                                               metadata_only=metadata_only, index_entries_only=index_entries_only)
        response = self._requests_handler.http_request_handler(path, "GET",
                                                               force_read_from_master=force_read_from_master).json()
        if "Error" in response:
            raise exceptions.ErrorResponseException(response["Error"][:100])
        return response

    def stream_query(self, index_name, index_query):
        """
        @param index_name: A name of an index to query
        :type str
        @param index_query: A query definition containing all information required to query a specified index..
        :type IndexQuery
        """
        if not index_name:
            raise ValueError("index_name cannot be None or empty")
        if index_query is None:
            raise ValueError("None query is invalid")
        if not isinstance(index_query, IndexQuery):
            raise ValueError("index_query must be IndexQuery type")
        method = "GET"
        if index_query.query and len(index_query.query) > 1024 + 512:
            method = "POST"

        path = "streams/query/{0}?".format(index_name)
        data = None
        headers = None
        if method == "GET":
            path += self._build_query_request_path(path=path, index_query=index_query, add_page_size=False)
        else:
            path += self._build_query_request_path(path=path, index_query=index_query, include_query=False,
                                                   add_page_size=False)
            data = index_query.query
            headers = {'content-encoding': 'gzip'}

        return self._requests_handler.http_request_handler(path=path, method=method, data=data, headers=headers,
                                                           stream=True)

    # For Admin use only (create or delete databases)
    class Admin(object):
        def __init__(self, requests_handler):
            self.requests_handler = requests_handler

        def create_database(self, database_document):
            """
            Creates a database

            @param database_document: has to be DatabaseDocument type
            """
            if "Raven/DataDir" not in database_document.settings:
                raise exceptions.InvalidOperationException("The Raven/DataDir setting is mandatory")
            db_name = database_document.database_id.replace("Raven/Databases/", "")
            Utils.name_validation(db_name)
            path = "databases/{0}".format(Utils.quote_key(db_name))
            response = self.requests_handler.http_request_handler(path, "PUT", database_document.to_json(),
                                                                  admin=True)

            if response.status_code == 502:
                raise exceptions.ErrorResponseException(
                    "Connection failed please check your connection to {0}".format(self.requests_handler.url))
            if response.status_code != 200:
                raise exceptions.ErrorResponseException(
                    "Database with the name '{0}' already exists".format(database_document.database_id))

            return response

        def delete_database(self, db_name, hard_delete=False):
            db_name = db_name.replace("Rave/Databases/", "")
            path = "databases/{0}".format(Utils.quote_key(db_name))
            if hard_delete:
                path += "?hard-delete=true"
            response = self.requests_handler.http_request_handler(path, "DELETE", admin=True)
            if response.content != '' and response.content != b'':
                raise exceptions.ErrorResponseException(response.content)
            return response

        def get_store_statistics(self):
            response = self.requests_handler.http_request_handler("stats", "GET", admin=True)
            if response.status_code == 200:
                return response.json()
